import os, socket
import Error
from ftplib import FTP, FTP_PORT, error_reply, error_perm		# FTP library
from Constants import FALSE, TRUE								# Boolean Values
from Constants import CS_OFFLINE, CS_ONLINE						# Connection Status
from Constants import CT_SSH, CT_RSH, CT_FTP, CT_LOCAL			# Connection Types

# ------------------------------------------------------------------------------
# Connection Descriptor

class Connection:

	def __init__(self, protocol = None):
		self.protocol	= protocol
		self.status		= CS_OFFLINE

	def login(self):
		self.status = CS_ONLINE

	def logoff(self):
		self.status = CS_OFFLINE
	
	def isonline(self):
		return (self.status == CS_ONLINE)

	def isoffline(self):
		return (self.status == CS_OFFLINE)

# ------------------------------------------------------------------------------
# Connection to a local host

class LocalConnection(Connection):

	def __init__(self):
		self.protocol	= CT_LOCAL
		self.status		= CS_ONLINE				# Assumes local FS is active

# ------------------------------------------------------------------------------
# Description of a connection to a remote host

class RemoteConnection(Connection):

	def __init__(self, protocol = None):
		self.protocol	= protocol				# Protocol to use
		self.hostname	= None					# Name of the remote host
		self.port		= None					# Port to use
		self.username	= None					# Login name
		self.password	= None					# Password to use
		self.status		= CS_OFFLINE			# Connection status (Inactive)

# ------------------------------------------------------------------------------
# Class holding features related to the FTP server

class FTPServerFeatures:

	def __init__(self):
		# Features an FTP server may support
		self.have_mdtm			= FALSE			# 'MDTM' commands (get last modification time)
		self.have_site_utime	= FALSE			# 'SITE UTIME' commands (set last modification time)
		self.have_otf_tar		= FALSE			# On-The-Fly archiving of files (.tar)
		self.have_otf_compress	= FALSE			# On-The-Fly compression of files (.Z)
		self.have_otf_gzip		= FALSE			# On-The-Fly compression of files (.gz)

		# Have we checked for those features yet ?
		self.checked_mdtm		= FALSE
		self.checked_site_utime	= FALSE
		self.checked_otf_tar	= FALSE
		self.checked_otf_compr	= FALSE
		self.checked_otf_gzip	= FALSE

		# Extra data used if above features are supported
		self.gmt_offset			= -1			# GTM - Local Time (in seconds)

# ------------------------------------------------------------------------------
# Connection Descriptor for the FTP protocol

class FTPConnection(RemoteConnection):

	def __init__(self, hostname = None, username = "anonymous", password = ""):
		self.protocol	= CT_FTP
		self.hostname	= hostname
		self.port		= FTP_PORT
		self.username	= username
		self.password	= password
		self.ftp		= None
		self.status		= CS_OFFLINE
		self.features	= FTPServerFeatures()

	def check_features_for_reading(self, path):
		if self.isonline():
			# Check whether the remote host supports 'MDTM' commands
			if not self.features.checked_mdtm:
				try:
					self.ftp.voidcmd("MDTM " + path)
				except error_perm, reply:
					self.features.have_mdtm = (str(reply)[:3] != "500")
				else:
					self.features.have_mdtm = TRUE
					self.features.checked_mdtm = TRUE

	def check_features_for_writing(self, path):
		pass

	def check_features(self, path):
		self.check_features_for_reading(path)
		self.check_features_for_writing(path)
		
	def login(self):
		if self.isonline():
			self.logoff()

		self.ftp = FTP()
		try:
			self.ftp.connect(self.hostname, self.port)
			self.ftp.login(self.username, self.password)
			# Set to passive mode in case we are behind a firewall
			self.ftp.set_pasv(TRUE)
		except socket.error, error_msg:
			raise Error.UnknownHost(self.hostname)
		except error_perm, error_msg:
			if (str(error_msg)[:3] == "530"):
				raise Error.LoginFailed()
			else:
				raise Error.FTPError(str(error_msg))
		except EOFError:
			raise Error.ConnectionClosed()
		else:
			self.status = CS_ONLINE
		
	def logoff(self):
		self.status = CS_OFFLINE
		if self.ftp:
			self.ftp.quit()

# ------------------------------------------------------------------------------
# Local Variables:
# tab-width: 4
# End:
