import os, re, sys, string

from stat import *
from time import strftime, mktime, gmtime, time, timezone				# Time utilities
from string import split, strip											# String utilities
from ftplib import FTP, error_reply, error_perm, error_proto			# FTP library
from Error import Error, InternalError, DirectoryNotFound				# Error Descriptors
from Connection import LocalConnection, FTPConnection					# Connection Descriptor
from Constants import TRUE, FALSE										# Boolean values
from Constants import PI_NAME, PI_VERSION, PI_LOCALE_DIR, PI_LOCALE_NAME# Program Information
from Constants import FT_DIRECTORY, FT_LINK, FT_REGULAR_FILE			# File Types
from Constants import CT_SSH, CT_RSH, CT_FTP, CT_LOCAL					# Connection Types
from Constants import RE_LISTDIR_UNIX									# FTP Directory Parsing RE
from Constants import MONTH_ID											# Month IDs   
from Constants import FTP_TIME_STAMP_FILE								# Exclude that file from LocalFileTree

# ------------------------------------------------------------------------------
# Localization

import gettext
gettext.bindtextdomain(PI_LOCALE_NAME, PI_LOCALE_DIR)
gettext.textdomain(PI_LOCALE_NAME)
_ = gettext.gettext

# ------------------------------------------------------------------------------
# File Descriptor

class File:
	
	def __init__(self, name = None, type = None):
		self.name		= name			# File name
		self.type		= type			# File type (Regular File, Directory, Link)
		self.size		= 0				# Size of the file
		self.permissions= 0				# File Permissions + sticky bit + set-{group,user}-id bits
		self.modtime	= 0				# Time of last modification
		self.linktarget	= ""			# Name of the target if (type == FT_LINK)
		self.contents	= []			# Contents of the directory (type == FT_DIRECTORY)

	def isfile(self):
		return self.type == FT_REGULAR_FILE
	
	def isdir(self):
		return self.type == FT_DIRECTORY
	
	def islink(self):
		return self.type == FT_LINK

# ------------------------------------------------------------------------------
# File Tree (Wrapper)

class FileTree:
	
	def __init__(self, connection):
		self.connection	= connection
		self.root_dir	= None
		self.file_tree	= None

	def __split_subdirs(self, path):
		split_basedir = split(self.file_tree.name, "/")
		split_currdir = split(path, "/")
		if (len(split_currdir) < len(split_basedir)):
			raise Error("FileTree::__split_subdirs", _("Depth of visited directory is lower than depth of root directory"))
		for i in range(len(split_basedir)):
			if (split_basedir[i] != split_currdir[i]):
				raise Error("FileTree::__split_subdirs", _("Visited directory has a different base than the root directory (%s:%s)") % (path, self.file_tree.name))
		dir_base = split_currdir[0:len(split_basedir)]
		dir_diff = split_currdir[len(split_basedir):]
		return dir_base, dir_diff
	
	def get_file_tree_node(self, path):
		dir_base, dir_diff = self.__split_subdirs(path)
		current_node = self.file_tree
		current_path = self.file_tree.name
		while dir_diff:
			current_dirname = dir_diff.pop(0)
			current_path += "/" + current_dirname
			found_dir = FALSE
			for entry in current_node.contents:
				if entry.isdir() and (entry.name == current_dirname):
					found_dir = TRUE
					current_node = entry
					break
			if not found_dir:
				raise DirectoryNotFound(current_path)
		return current_node

	def __print_file_tree_cb(self, path, file_entry, arg):
		full_name = os.path.normpath(os.path.join(path, file_entry.name))
		if file_entry.isdir():
			print "* Directory: %s" % full_name
			print "  Contents: %d entries" % len(file_entry.contents)
		else:
			print "- File: %s" % full_name
			if file_entry.islink():
				print "  Link to: %s" % file_entry.linktarget
			else:
				print "  Date: %s" % strftime("%Y/%m/%d - %H:%M:%S", gmtime(file_entry.modtime))
				print "  Size: %d bytes" % file_entry.size
		
	def getTree(self, root_dir = "/"):
		self.root_dir = os.path.normpath(root_dir)
		if (self.connection.protocol == CT_SSH) or (self.connection.protocol == CT_RSH):
			# Do nothing because we don't need to fetch the source and dest
			# file lists with those protocols (rsync does the job for us)
			self.file_tree = File(self.root_dir, FT_DIRECTORY)
		elif (self.connection.protocol == CT_LOCAL):
			self.file_tree = LocalFileTree(self.connection).getTree(self.root_dir)
		elif (self.connection.protocol == CT_FTP):
			self.file_tree = FTPFileTree(self.connection).getTree(self.root_dir)
		else:
			raise Error("FileTree::getTree", _("Unknown connection type"))
		return self.file_tree

	def printTree(self):
		self.traverse(self.__print_file_tree_cb, None)

	def __traverse_rec(self, path, file_tree, callback, arg):
		norm_path = os.path.normpath(path)
		for entry in file_tree.contents:
			callback(norm_path, entry, arg)
			if entry.isdir():
				new_path = os.path.join(norm_path, entry.name)
				self.__traverse_rec(new_path, entry, callback, arg)
	
	def traverse(self, callback, arg):
		self.__traverse_rec(self.file_tree.name, self.file_tree, callback, arg)

# ------------------------------------------------------------------------------
# File Tree (FTP)

class FTPFileTree(FileTree):
	
	def __init__(self, connection):
		self.connection		= connection
		self.root_dir		= None
		self.file_tree		= None
		self.ftp			= None
		self.parse_list_dir	= None
		self.compiled_regex	= None
		self.current_year	= gmtime(time())[0]

	def __get_file_type(self, file_type_str):
		if (file_type_str == "-"):
			file_type = FT_REGULAR_FILE
		elif (file_type_str == "d"):
			file_type = FT_DIRECTORY
		elif (file_type_str == "l"):
			file_type = FT_LINK
		return file_type
	
	def __get_file_permissions(self, permissions_str):
		permissions = 0
		
		# Owner permissions
		if (permissions_str[0] == "r"):
			permissions |= S_IRUSR
		if (permissions_str[1] == "w"):
			permissions |= S_IWUSR
		if (permissions_str[2] == "x"):
			permissions |= S_IXUSR

		# Group permissions
		if (permissions_str[3] == "r"):
			permissions |= S_IRGRP
		if (permissions_str[4] == "w"):
			permissions |= S_IWGRP
		if (permissions_str[5] == "x"):
			permissions |= S_IXGRP
		
		# Other permissions
		if (permissions_str[6] == "r"):
			permissions |= S_IROTH
		if (permissions_str[7] == "w"):
			permissions |= S_IWOTH
		if (permissions_str[8] == "x"):
			permissions |= S_IXOTH
		
		return permissions

	def __parse_ls_date(self, day, month, year_or_hour):
		day		= int(day)
		month	= MONTH_ID[string.lower(month)]
		if (year_or_hour[2:3] == ":"):
			year		= self.current_year
			hours		= int(year_or_hour[:2])
			minutes		= int(year_or_hour[3:])
		else:
			year		= int(year_or_hour)
			hours		= 0
			minutes		= 0
		return int(mktime((year, month, day, hours, minutes, 0, 0, 0, -1)))
	
	def __parse_list_dir_unix(self, line):
		# match 	= re.match(RE_LISTDIR_UNIX, line, re.IGNORECASE)
		match = self.compiled_regex.match(line)
		if match:
			new_file			= File()
			new_file.type		= self.__get_file_type(match.group(1))
			new_file.permissions= self.__get_file_permissions(match.group(2))
			new_file.size		= int(match.group(6))
			new_file.name		= match.group(10)
			new_file.modtime	= self.__parse_ls_date(match.group(8), match.group(7), match.group(9))
			new_file.contents	= []

			if new_file.islink():
				split_name = split(new_file.name, "->")
				new_file.name = strip(split_name[0])
				new_file.linktarget = strip(split_name[1])

			if (new_file.name == ".") or (new_file.name == ".."):
				match = None

		if match:
			return new_file
		else:
			return None

	def __get_file_modtime(self, path, old_modtime):
		modtime = old_modtime
		if self.connection.features.have_mdtm:
			# The time returned by "MDTM" is relative to GMT
			reply = self.ftp.sendcmd("MDTM " + os.path.normpath(path))
			if (reply[:3] != "550"):
				modtime	= reply[4:]
				year	= int(modtime[0:4])
				month	= int(modtime[4:6])
				day		= int(modtime[6:8])
				hours	= int(modtime[8:10])
				minutes	= int(modtime[10:12])
				seconds	= int(modtime[12:14])
				# Compute time against Epoch and adjust according to timezone
				modtime	= int(mktime((year, month, day, hours, minutes, seconds, 0, 0, 0)))
				modtime = modtime - timezone
		return modtime
	
	def __get_file_tree_rec(self, path):
		full_path = os.path.normpath(path)
		
		list_lines = []
		self.ftp.dir(full_path, list_lines.append)
		dir_node = self.get_file_tree_node(full_path)

		for line in list_lines:
			new_file = self.parse_list_dir(line)
			if (new_file != None):
				full_name = os.path.join(full_path, new_file.name)
				dir_node.contents.append(new_file)
				if new_file.isdir():
					# Recurse through sub directories
					self.__get_file_tree_rec(full_name)
				elif new_file.isfile():
					# Fix last modification time info thanks to 'MDTM' command, if supported
					new_file.modtime = self.__get_file_modtime(full_name, new_file.modtime)

	def __setup_list_dir_parser_cb(self, line):
		if (self.parse_list_dir == None):
			# There are (at least?) two formats for Unix ftp servers
			# -rw-r--r--   1 14       50           4399 Jul  1  2000 TODO
			# drwxr-xr-x   2 14       50           4096 Mar  6 16:58 examples
			if (re.match(RE_LISTDIR_UNIX, line, re.IGNORECASE) != None):
				self.parse_list_dir = self.__parse_list_dir_unix
				self.compiled_regex = re.compile(RE_LISTDIR_UNIX, re.IGNORECASE)

	def __setup_list_dir_parser(self, test_dir):
		self.ftp.dir(test_dir, self.__setup_list_dir_parser_cb)
		if (self.parse_list_dir == None):
			# FIXME: Assume remote directory was empty and remote fs type is unix
			self.parse_list_dir = self.__parse_list_dir_unix
			self.compiled_regex = re.compile(RE_LISTDIR_UNIX, re.IGNORECASE)

	def getTree(self, root_dir = "/"):
		self.root_dir = os.path.normpath(root_dir)
		self.connection.login()
		self.connection.check_features_for_reading(self.root_dir)

		try:
			self.ftp = self.connection.ftp
			self.ftp.cwd(self.root_dir)
		except error_perm, message:
			if (str(message)[:3] == "550"):
				raise DirectoryNotFound(self.root_dir)
		
		self.__setup_list_dir_parser(self.root_dir)

		# Do actual recursion through sub directories
		self.file_tree = File(self.root_dir, FT_DIRECTORY)
		self.file_tree.contents = []
		self.__get_file_tree_rec(self.root_dir)

		self.connection.logoff()
		return self.file_tree

# ------------------------------------------------------------------------------
# File Tree (Local)

class LocalFileTree(FileTree):

	def __init__(self, connection):
		self.connection	= connection
		self.root_dir	= None
		self.file_tree	= None
	
	def __get_file_permissions(self, path):
		st_mode = os.stat(path)[ST_MODE]
		return S_IMODE(st_mode)
	
	def __get_file_modtime(self, path):
		st_mtime = os.stat(path)[ST_MTIME]
		# Force 'seconds' field to be zero and remain consistent with FTPFileTree
		# st_mtime = st_mtime - (st_mtime % 60)
		return st_mtime
	
	def __get_file_size(self, path):
		st_size = os.stat(path)[ST_SIZE]
		return st_size
	
	def __get_file_tree_cb(self, __unusued_arg,  path, names):
		dir_node = self.get_file_tree_node(path)
		for filename in names:
			if filename not in [ FTP_TIME_STAMP_FILE ]:
				full_filename = path + "/" + filename
				new_file = File(filename)
				new_file.permissions = self.__get_file_permissions(full_filename)
				new_file.modtime = self.__get_file_modtime(full_filename)
				new_file.contents = []
				if os.path.islink(full_filename):
					new_file.type = FT_LINK
					new_file.linktarget = os.readlink(full_filename)
				elif os.path.isdir(full_filename):
					new_file.type = FT_DIRECTORY
				elif os.path.isfile(full_filename):
					new_file.type = FT_REGULAR_FILE
					new_file.size = self.__get_file_size(full_filename)
				else:
					raise Error("LocalFileTree::__get_file_tree_cb", _("Unknown file type"))
				dir_node.contents.append(new_file)
	
	def getTree(self, root_dir = "/"):
		self.root_dir = os.path.normpath(root_dir)
		
		# Root directory shall be a directory
		if not os.path.isdir(self.root_dir):
			raise DirectoryNotFound(self.root_dir)

		self.file_tree				= File(self.root_dir, FT_DIRECTORY)
		self.file_tree.permissions	= self.__get_file_permissions(self.root_dir)
		self.file_tree.modtime		= self.__get_file_modtime(self.root_dir)
		self.file_tree.contents		= []
		os.path.walk(self.root_dir, self.__get_file_tree_cb, None)
		return self.file_tree

# ------------------------------------------------------------------------------
# Testers

def test_local(args):
	path = args[0]
	connection = LocalConnection()
	file_tree = FileTree(connection)
	file_tree.getTree(path)
	file_tree.printTree()

def get_login():
	sys.stdout.write("Username: ")
	username = raw_input()
	if (username == ""):
		username = "anonymous"
	sys.stdout.write("Password: ")
	password = raw_input()
	return username, password
		
def test_ftp(args):
	hostname = args[0]
	path = args[1]
	username, password = get_login()
	connection = FTPConnection(hostname, username, password)
	file_tree = FileTree(connection)
	file_tree.getTree(path)
	file_tree.printTree()

if __name__ == "__main__":
	test = sys.argv[1]
	if (test == "test_local"):
		test_local(sys.argv[2:])
	elif (test == "test_ftp"):
		test_ftp(sys.argv[2:])
	else:
		print "Error: invalid test module: %s" % test

# ------------------------------------------------------------------------------
# Local Variables:
# tab-width: 4
# End:
