import os, sys
import Error

from ftplib import FTP, FTP_PORT, error_reply, error_perm				# FTP library
from time import strftime, gmtime, time									# Time Utilities
from FileTree import File, FTPFileTree, LocalFileTree					# File Trees
from Connection import LocalConnection, FTPConnection					# Connection Types
from Error import Error, InternalError, FTPError						# Error Descriptors
from Constants import TRUE, FALSE										# Boolean values
from Constants import PI_NAME, PI_LOCALE_DIR, PI_LOCALE_NAME			# Program Information
from Constants import CT_SSH, CT_RSH, CT_FTP, CT_LOCAL					# Connection Types
from Constants import *													# Hmmm, finally import all constants and support functions

# ------------------------------------------------------------------------------
# Localization

import gettext
gettext.bindtextdomain(PI_LOCALE_NAME, PI_LOCALE_DIR)
gettext.textdomain(PI_LOCALE_NAME)
_ = gettext.gettext

# ------------------------------------------------------------------------------
# Descriptor of a file to synchronize

class SyncFile(File):

	def __init__(self, file, command = None, copy_subdirs = FALSE):
		self.name		= file.name
		self.type		= file.type
		self.size		= file.size
		self.permissions= file.permissions
		self.modtime	= file.modtime
		self.linktarget	= file.linktarget
		self.command	= command
		self.contents	= []
		if copy_subdirs:
			self.__copy_file_tree_rec(self.contents, file, command)

	def __copy_file_tree_rec(self, sync_file_tree, file_tree, command):
		# Note: when copying sub directories, command is expected to be consistent
		# i.e. this is likely to be an insertion of a complete directory into
		# either the LOCAL or REMOTE directory
		for entry in file_tree.contents:
			sync_file = SyncFile(entry, command)
			if sync_file.isdir():
				self.__copy_file_tree_rec(sync_file.contents, entry, command)
			sync_file_tree.append(sync_file)

# ------------------------------------------------------------------------------
# Progress Wrapper for file descriptors
# (enables updates of the GUI progress bar)

class FileProgressWrapper:
	def __init__(self, file, file_size, block_size, callback):
		self.file		= file
		self.callback	= callback
		self.curr_size	= 0
		self.block_size	= block_size
		if file_size < block_size:
			self.file_size = block_size
		else:
			self.file_size	= file_size

	def write(self, block):
		self.file.write(block)
		self.curr_size = self.curr_size + len(block)
		self.callback(float(self.curr_size) / float(self.file_size))

	def read(self, block_size):
		block = self.file.read(block_size)
		self.curr_size = self.curr_size + len(block)
		self.callback(float(self.curr_size) / float(self.file_size))
		return block

	def close(self):
		self.file.close()
	
# ------------------------------------------------------------------------------
# Synchronized File Tree

class SyncFileTree:

	def __init__(self, sync_type = None, local_file_tree = None, remote_file_tree = None):
		# Note: if both connections are LOCAL, the LOCAL item is the source
		# directory and the REMOTE item is the target directory
		self.sync_type			= sync_type
		self.local_connection	= local_file_tree.connection
		self.local_file_tree	= local_file_tree
		self.local_root_dir		= local_file_tree.root_dir
		self.remote_connection	= remote_file_tree.connection
		self.remote_file_tree	= remote_file_tree
		self.remote_root_dir	= remote_file_tree.root_dir
		self.local_file_tree	= local_file_tree
		self.remote_file_tree	= remote_file_tree
		self.sync_file_tree		= []
		self.current_year		= gmtime(time())[0]
		self.do_synchronize		= None
		self.BLOCK_SIZE			= 1024

	def __different_modtime_large(self, x_modtime, y_modtime):
		# Large difference computation. The rules depend on the relative 'age' of the entries
		# 1. The files are old enough: it is sufficient to rely on yyyy/mm/dd
		# 2. The files are new enough: try to rule out with higher precision (seconds)
		x_time = gmtime(x_modtime)
		y_time = gmtime(y_modtime)
		
		# Common case: compare against year, month and day
		are_different = (x_time[0] != y_time[0]) or (x_time[1] != y_time[1]) or (x_time[2] != y_time[2])
		if (x_time[0] != self.current_year) or (y_time[0] != self.current_year):
			# Entries are old enough for the common case to be valid
			pass
		else:
			# Try to rely on greater precision: hour, minutes, (seconds?)
			are_different = are_different or (x_time[3] != y_time[3]) or (x_time[4] != y_time[4])
		return are_different
		
	def __traverse_differences_rec(self, callback, sync_file_tree, local_file_tree, remote_file_tree):
		# Deal only with the contents of the directories
		local_dir_contents = local_file_tree.contents
		remote_dir_contents = remote_file_tree.contents

		if not len(local_dir_contents) and not len(remote_dir_contents):
			# Pass if both directories are empty
			pass
		else:
			# Check for items in local but not in remote
			new_local_entries = [x for x in local_dir_contents if x.name not in map(lambda y: y.name, remote_dir_contents)]
			for entry in new_local_entries:
				callback(sync_file_tree, entry, None)

			# Check for items in remote but not in local
			new_remote_entries = [x for x in remote_dir_contents if x.name not in map(lambda y: y.name, local_dir_contents)]
			for entry in new_remote_entries:
				callback(sync_file_tree, None, entry)

			# Check for items with the same name. There are several cases:
			# 1. Types differ
			#    -> mark the items as different
			# 2. Types are identical:
			#    * item is a directory:
			#      -> check sub directories
			#    * item is a link:
			#      -> check link target
			#    * item is a regular file:
			#      -> check last modification time, size and permissions
			different_entries = [(x, y) for x in local_dir_contents for y in remote_dir_contents if
								 ((x.name == y.name)
								  and ((x.type != y.type)
									   or (x.isdir())
									   or (x.isfile() and (self.__different_modtime_large(x.modtime,  y.modtime) or (x.size != y.size) or (x.permissions != y.permissions)))
									   or (x.islink() and (x.linktarget != y.linktarget))))]
			for (local_entry, remote_entry) in different_entries:
				if (local_entry.isdir() and remote_entry.isdir()):
					# Both items are directories, recursive sub directories
					sync_entry = SyncFile(local_entry, SC_TRAVERSE_DIRECTORY, FALSE)
					sync_file_tree.append(sync_entry)
					self.__traverse_differences_rec(callback, sync_entry.contents, local_entry, remote_entry)
				else:
					# Append the item according to the synchronization type
					callback(sync_file_tree, local_entry, remote_entry)

	def __traverse_differences(self, callback):
		self.__traverse_differences_rec(callback, self.sync_file_tree, self.local_file_tree.file_tree, self.remote_file_tree.file_tree)

	def __prepare_local_overwrites_remote_cb(self, sync_file_tree, local_file, remote_file):
		append_new_file = TRUE
		# sync_file_tree is a list (directory) of files
		files = (local_file, remote_file)
		if (files == (None, None)):
			# Fail if both items are None
			raise InternalError("SyncFileTree::__prepare_local_overwrites_remote_cb", "Both items are nil")
		elif (files == (local_file, None)):
			# Recursively copy the local directory to the remote directory
			new_file = SyncFile(local_file, SC_INSERT_REMOTE, TRUE)
		elif (files == (None, remote_file)):
			# Recursively remove the remote directory
			new_file = SyncFile(remote_file, SC_REMOVE_REMOTE, TRUE)
		elif (local_file.modtime > self.last_sync):
			# Mark the remote file as to be updated by the local file
			new_file = SyncFile(local_file, SC_UPDATE_REMOTE, FALSE)
		else:
			# Nothing new since last successful synchronization
			append_new_file = FALSE
		if append_new_file:
			sync_file_tree.append(new_file)

	def __prepare_remote_overwrites_local_cb(self, sync_file_tree, local_file, remote_file):
		append_new_file = TRUE
		# sync_file_tree is a list (directory) of files
		files = (local_file, remote_file)
		if (files == (None, None)):
			# Fail if both items are None
			raise InternalError("SyncFileTree::__prepare_remote_overwrites_local_cb", "Both items are nil")
		elif (files == (local_file, None)):
			# Recursively remote the local directory
			new_file = SyncFile(local_file, SC_REMOVE_LOCAL, TRUE)
		elif (files == (None, remote_file)):
			# Recursively copy the remote directory to the local directory
			new_file = SyncFile(remote_file, SC_INSERT_LOCAL, TRUE)
		elif (remote_file.modtime > self.last_sync):
			# Mark the local file as to be updated by the remote file
			new_file = SyncFile(remote_file, SC_UPDATE_LOCAL, FALSE)
		else:
			# Nothing new since last successful synchronization
			append_new_file = FALSE
		if append_new_file:
			sync_file_tree.append(new_file)

	def __prepare_newer_overwrites_older_cb(self, sync_file_tree, local_file, remote_file):
		# sync_file_tree is a list (directory) of files
		files = (local_file, remote_file)
		if (files == (None, None)):
			# Fail if both items are None
			raise InternalError("SyncFileTree::__prepare_newer_overwrites_older_cb", "Both items are nil")
		elif (files == (local_file, None)):
			# Recursively copy the local file (possibly dir) to remote
			# FIXME: do we consider the modtime of an inexistent dir/file as -oo ?
			new_file = SyncFile(local_file, SC_INSERT_REMOTE, TRUE)
		elif (files == (None, remote_file)):
			# Recursively copy the remote file (possibly dir) to local
			# FIXME: do we consider the modtime of an inexistent dir/file as -oo ?
			new_file = SyncFile(remote_file, SC_INSERT_LOCAL, TRUE)
		elif (local_file.modtime > remote_file.modtime):
			# Update remote file
			new_file = SyncFile(local_file, SC_UPDATE_REMOTE, FALSE)
		elif (local_file.modtime < remote_file.modtime):
			# Update local file
			new_file = SyncFile(remote_file, SC_UPDATE_LOCAL, FALSE)
		elif (local_file.permissions != remote_file.permissions):
			# Error to be caught: different file permissions
			# NOTE: both items can't be directories at the same time
			raise Error.MutualSyncDifferentPermissions()
		elif (local_file.size != remote_file.size):
			# Error to be caught: different file size
			# NOTE: both items can't be directories at the same time
			raise Error.MutualSyncDifferentSize()
		else:
			raise Error.InternalError("SyncFileTree::__prepare_newer_overwrites_older_cb", "unexpected error")
		sync_file_tree.append(new_file)
	
	def prepare(self, last_sync, sync_type = None):
		self.last_sync = last_sync
		if (sync_type != None):
			self.sync_type = sync_type
		if (self.sync_type == ST_LOCAL_IS_NEWER):
			self.__traverse_differences(self.__prepare_local_overwrites_remote_cb)
		elif (self.sync_type == ST_REMOTE_IS_NEWER):
			self.__traverse_differences(self.__prepare_remote_overwrites_local_cb)
		elif (self.sync_type == ST_EITHER_IS_NEWER):
			self.__traverse_differences(self.__prepare_newer_overwrites_older_cb)
		else:
			raise Error("SyncFileTree::prepare", _("Unknown type of synchronization"))

	def __traverse_rec(self, path, sync_dir_contents, callback, arg):
		for entry in sync_dir_contents:
			callback(path, entry, arg)
			if self.stop_traversal:
				break
			if entry.isdir():
				new_path = os.path.normpath(os.path.join(path, entry.name))
				self.__traverse_rec(new_path, entry.contents, callback, arg)

	def traverse(self, callback, arg):
		# Argument format of callback function: callback(path, sync_file, arg)
		self.stop_traversal = FALSE
		self.__traverse_rec("", self.sync_file_tree, callback, arg)

	def __print_entry(self, path, sync_file, __unused_arg):
		full_name = os.path.normpath(os.path.join(path, sync_file.name))
		if sync_file.isdir():
			print "* Directory: %s" % full_name
			print "  Contents: %d entries" % len(sync_file.contents)
		else:
			print "- File: %s" % full_name
			if sync_file.islink():
				print "  Link to: %s" % sync_file.linktarget
			else:
				print "  Date: %s" % strftime("%Y/%m/%d - %H:%M:%S", gmtime(sync_file.modtime))
				print "  Size: %d bytes" % sync_file.size
		print "  Action: %s" % SC_STRINGS[sync_file.command]
	
	def printTree(self):
		self.traverse(self.__print_entry, None)

	def __copy_local_to_local(self, sync_file, source_filename, target_filename):
		if sync_file.isfile():
			source_file = open(source_filename, "rb")
			target_file = open(target_filename, "wb")

			while TRUE:
				block = source_file.read(self.BLOCK_SIZE)
				if (block == ""):
					break
				target_file.write(block)

			source_file.close()
			target_file.close()
			
			# Permissions and time-stamp
			os.chmod(target_filename, sync_file.permissions)
			os.utime(target_filename, (sync_file.modtime, sync_file.modtime))

		elif sync_file.isdir():
			os.mkdir(target_filename, sync_file.permissions)
			source_dirname = source_filename
			target_dirname = target_filename
			for file in sync_file.contents:
				source_filename = os.path.normpath(os.path.join(source_dirname, file.name))
				target_filename = os.path.normpath(os.path.join(target_dirname, file.name))
				self.__copy_local_to_local(file, source_filename, target_filename)
			sync_file.contents = []

	def __remove_local(self, sync_file, filename):
		if sync_file.isfile():
			os.remove(filename)

		elif sync_file.isdir():
			dirname = filename
			for file in sync_file.contents:
				filename = os.path.normpath(os.path.join(dirname, file.name))
				self.__remove_local(file, filename)
			os.rmdir(dirname)
			sync_file.contents = []
	
	def __do_synchronize_local_local(self, path, sync_file, transfer):
		command = sync_file.command
		if is_local_command(command):
			source_filename = os.path.normpath(os.path.join(self.remote_root_dir, path, sync_file.name))
			target_filename = os.path.normpath(os.path.join(self.local_root_dir, path, sync_file.name))
		elif is_remote_command(command):
			source_filename = os.path.normpath(os.path.join(self.local_root_dir, path, sync_file.name))
			target_filename = os.path.normpath(os.path.join(self.remote_root_dir, path, sync_file.name))

		command = command_id(command)
		if (command == SC_INSERT) or (command == SC_UPDATE):
			self.__copy_local_to_local(sync_file, source_filename, target_filename)
		elif (command == SC_REMOVE):
			self.__remove_local(sync_file, target_filename)

	def __copy_ftp_to_local(self, sync_file, remote_filename, local_filename):
		if self.stop_traversal:
			return
			
		if sync_file.isfile():
			# TODO: on-the-fly compression
			self.stop_traversal = self.gui_update_file_entry(sync_file.name)
			local_file = FileProgressWrapper(open(local_filename, "wb"), sync_file.size, self.BLOCK_SIZE, self.gui_update_file_progress)
			self.remote_connection.ftp.retrbinary("RETR %s" % remote_filename, local_file.write, self.BLOCK_SIZE)
			local_file.close()
			os.chmod(local_filename, sync_file.permissions)
			os.utime(local_filename, (sync_file.modtime, sync_file.modtime))
			self.gui_update_files_progress()

		elif sync_file.isdir():
			# TODO: on-the-fly tar'ing
			remote_dirname = remote_filename
			local_dirname  = local_filename
			self.gui_update_dir_entries(local_dirname, remote_dirname, sync_file.name)
			os.mkdir(local_dirname, sync_file.permissions)
			for file in sync_file.contents:
				remote_filename = os.path.normpath(os.path.join(remote_dirname, file.name))
				local_filename  = os.path.normpath(os.path.join(local_dirname , file.name))
				self.__copy_ftp_to_local(file, remote_filename, local_filename)
			sync_file.contents = []

	def __copy_local_to_ftp(self, sync_file, local_filename, remote_filename):
		if self.stop_traversal:
			return
		
		if sync_file.isfile():
			self.stop_traversal = self.gui_update_file_entry(sync_file.name)
			local_file = FileProgressWrapper(open(local_filename, "rb"), sync_file.size, self.BLOCK_SIZE, self.gui_update_file_progress)
			self.remote_connection.ftp.storbinary("STOR %s" % remote_filename, local_file, self.BLOCK_SIZE)
			local_file.close()
			# TODO: try to change modtime and permissions
			self.gui_update_files_progress()

		elif sync_file.isdir():
			remote_dirname = remote_filename
			local_dirname  = local_filename
			self.gui_update_dir_entries(local_dirname, remote_dirname, sync_file.name)
			self.remote_connection.ftp.mkd(remote_dirname)
			for file in sync_file.contents:
				remote_filename = os.path.normpath(os.path.join(remote_dirname, file.name))
				local_filename  = os.path.normpath(os.path.join(local_dirname , file.name))
				self.__copy_local_to_ftp(file, local_filename, remote_filename)
			sync_file.contents = []
	
	def __remove_ftp(self, sync_file, filename):
		if self.stop_traversal:
			return
		
		if sync_file.isfile():
			self.stop_traversal = self.gui_update_file_entry(sync_file.name)
			if self.delete_files:
				# gb-- this is safe to test only delete_files because the case "newer overwrites older" doesn't remove anything
				self.remote_connection.ftp.delete(filename)
					
			self.gui_update_files_progress()
		
		elif sync_file.isdir():
			dirname = filename
			for file in sync_file.contents:
				filename = os.path.normpath(os.path.join(dirname, file.name))
				self.__remove_ftp(file, filename)
			self.remote_connection.ftp.rmd(dirname)
			sync_file.contents = []
	
	def __do_synchronize_local_ftp(self, path, sync_file, args):
		local_dirname	= os.path.normpath(os.path.join(self.local_root_dir, path))
		local_filename	= os.path.normpath(os.path.join(local_dirname, sync_file.name))
		remote_dirname	= os.path.normpath(os.path.join(self.remote_root_dir, path))
		remote_filename	= os.path.normpath(os.path.join(remote_dirname, sync_file.name))

		command = sync_file.command
		if (command == SC_UPDATE_LOCAL) or (command == SC_INSERT_LOCAL):
			self.__copy_ftp_to_local(sync_file, remote_filename, local_filename)
		elif (command == SC_UPDATE_REMOTE) or (command == SC_INSERT_REMOTE):
			self.__copy_local_to_ftp(sync_file, local_filename, remote_filename)
		elif (command == SC_REMOVE_LOCAL):
				self.__remove_local(sync_file, local_filename)
		elif (command == SC_REMOVE_REMOTE):
				self.__remove_ftp(sync_file, remote_filename)

	def synchronize(self, args = None):
		# Setup hooks
		if (self.remote_connection.protocol == CT_FTP):
			self.do_synchronize = self.__do_synchronize_local_ftp
		elif (self.local_connection.protocol == CT_LOCAL) and (self.remote_connection.protocol == CT_LOCAL):
			self.do_synchronize = self.__do_synchronize_local_local

		# Sanity checks
		if (self.do_synchronize == None):
			raise InternalError("SyncFileTree::synchronize", "Invalid direction of synchronization")

		# Setup callbacks
		if args:
			self.gui_update_dir_entries		= args[0]
			self.gui_update_file_entry		= args[1]
			self.gui_update_file_progress	= args[2]
			self.gui_update_files_progress	= args[3]
			self.delete_files				= args[4]
			self.last_sync					= args[5]
		
		# Do the synchronization
		if (self.remote_connection.protocol == CT_FTP):
			self.remote_connection.login()

		try:
			self.traverse(self.do_synchronize, None)
		except error_perm, error_msg:
			raise FTPError(str(error_msg))

		if (self.remote_connection.protocol == CT_FTP):
			self.remote_connection.logoff()

# ------------------------------------------------------------------------------
# Testers

TEST_PREPARE_LOCAL	= "prepare_local"
TEST_PREPARE_FTP	= "prepare_ftp"
TEST_SYNC_LOCAL		= "sync_local"
TEST_SYNC_FTP		= "sync_ftp"

def print_usage_test_prepare_local():
	print "Usage: %s %s <( update_remote | update_local | update_oldest )> <local_dir> <remote_dir>" % (sys.argv[0], TEST_PREPARE_LOCAL)

def print_usage_test_prepare_ftp():
	print "Usage: %s %s <( update_remote | update_local | update_oldest )> <local_dir> <remote_dir> <remote_host>" % (sys.argv[0], TEST_PREPARE_FTP)

def do_test_prepare(sync_type, local_file_tree, remote_file_tree):
	if (sync_type == "update_remote"):
		sync_file_tree = SyncFileTree(ST_LOCAL_IS_NEWER, local_file_tree, remote_file_tree)
	elif (sync_type == "update_local"):
		sync_file_tree = SyncFileTree(ST_REMOTE_IS_NEWER, local_file_tree, remote_file_tree)
	elif (sync_type == "update_oldest"):
		sync_file_tree = SyncFileTree(ST_EITHER_IS_NEWER, local_file_tree, remote_file_tree)
	else:
		raise Error("do_test_prepare", "Invalid command (%s)" % sync_type)
	sync_file_tree.prepare()
	sync_file_tree.printTree()
	return sync_file_tree

def test_local(args, do_sync):
	if (len(args) != 3):
		print_usage_test_prepare_local()
	else:
		local_file_tree = LocalFileTree( LocalConnection() )
		local_file_tree.getTree(args[1])
		remote_file_tree = LocalFileTree( LocalConnection() )
		remote_file_tree.getTree(args[2])
		sync_file_tree = do_test_prepare(args[0], local_file_tree, remote_file_tree)
		if do_sync:
			sync_file_tree.synchronize()

def get_login():
	sys.stdout.write("Username: ")
	username = raw_input()
	if (username == ""):
		username = "anonymous"
	sys.stdout.write("Password: ")
	password = raw_input()
	return username, password
		
def test_ftp(args, do_sync):
	if (len(args) != 4):
		print_usage_test_prepare_ftp()
	else:
		local_file_tree = LocalFileTree( LocalConnection() )
		local_file_tree.getTree(args[1])
		username, password = get_login()
		remote_file_tree = FTPFileTree( FTPConnection(args[3], username, password) )
		remote_file_tree.getTree(args[2])
		sync_file_tree = do_test_prepare(args[0], local_file_tree, remote_file_tree)
		if do_sync:
			sync_file_tree.synchronize()

if __name__ == "__main__":
	if (len(sys.argv) < 2):
		print "Usage: %s <command> [<args>*]" % sys.argv[0]
	else:
		test = sys.argv[1]
		if (test == TEST_PREPARE_LOCAL) or (test == TEST_SYNC_LOCAL):
			test_local(sys.argv[2:], test == TEST_SYNC_LOCAL)
		elif (test == TEST_PREPARE_FTP) or (test == TEST_SYNC_FTP):
			test_ftp(sys.argv[2:], test == TEST_SYNC_FTP)
		else:
			print "Error: invalid test module: %s" % test

# ------------------------------------------------------------------------------
# Local Variables:
# tab-width: 4
# End:
